from django.shortcuts import render
from django.http import HttpResponse
from django.http import JsonResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseForbidden

from core.lights.ring import Ring
from core.lights.strip import Strip
from core.lights.screen import Screen
from core.lights.programs import Cava
from core.lights.programs import Disabled
from core.lights.programs import Fixed
from core.lights.programs import Rainbow
from core.lights.programs import Alarm
from core.lights.programs import Adaptive
from core.lights.circle.circle import Circle
from core.models import Setting
import core.state_handler as state_handler

from functools import wraps
import threading
import time


class Lights:
    def __init__(self, base):
        self.UPS = 30
        self.seconds_per_frame = 1 / self.UPS

        self.base = base
        self.ring = Ring()
        self.strip = Strip()
        self.screen = Screen()

        # if the led loop is running
        self.loop_active = threading.Event()

        self.cava_program = Cava(self)
        self.alarm_program = Alarm(self)
        # a dictionary containing all programs by their name
        self.programs = {}
        for program_class in [Disabled, Fixed, Alarm, Rainbow, Adaptive, Circle]:
            instance = program_class(self)
            self.programs[instance.name] = instance

        # this lock ensures that only one thread changes led options
        self.option_lock = threading.Lock()
        self.program_speed = 1

        last_ring_program_name = Setting.objects.get_or_create(
            key="last_ring_program", defaults={"value": "Disabled"}
        )[0].value
        last_strip_program_name = Setting.objects.get_or_create(
            key="last_strip_program", defaults={"value": "Disabled"}
        )[0].value
        last_screen_program_name = Setting.objects.get_or_create(
            key="last_screen_program", defaults={"value": "Disabled"}
        )[0].value
        ring_program_name = Setting.objects.get_or_create(
            key="ring_program", defaults={"value": "Disabled"}
        )[0].value
        strip_program_name = Setting.objects.get_or_create(
            key="strip_program", defaults={"value": "Disabled"}
        )[0].value
        screen_program_name = Setting.objects.get_or_create(
            key="screen_program", defaults={"value": "Disabled"}
        )[0].value

        self.last_ring_program = self.programs[last_ring_program_name]
        self.last_strip_program = self.programs[last_strip_program_name]
        self.last_screen_program = self.programs[last_screen_program_name]

        self.ring_program = self.programs[ring_program_name]
        self.strip_program = self.programs[strip_program_name]
        self.screen_program = self.programs[screen_program_name]
        # disable disconnected devices
        if not self.ring.initialized:
            self.ring_program = self.programs["Disabled"]
        if not self.strip.initialized:
            self.strip_program = self.programs["Disabled"]
        if not self.screen.initialized:
            self.screen_program = self.programs["Disabled"]
        self.ring_program.use()
        self.strip_program.use()
        self.screen_program.use()
        self.consumers_changed()

        self.start()

    def start(self):
        threading.Thread(target=self._loop, daemon=True).start()

    def _loop(self):
        iteration_count = 0
        adaptive_quality_window = self.UPS * 10
        time_sum = 0
        while True:
            self.loop_active.wait()

            with self.option_lock:
                computation_start = time.time()

                # these programs only actually compute something if they are used, so these functions are always called
                self.cava_program.compute()
                self.alarm_program.compute()

                if self.screen_program.name != "Disabled":
                    self.screen_program.draw()

                self.ring_program.compute()
                if self.strip_program != self.ring_program:
                    self.strip_program.compute()

                if self.ring_program.name != "Disabled":
                    if self.ring.monochrome:
                        ring_colors = [
                            self.ring_program.strip_color()
                            for led in range(self.ring.LED_COUNT)
                        ]
                    else:
                        ring_colors = self.ring_program.ring_colors()
                    self.ring.set_colors(ring_colors)

                if self.strip_program.name != "Disabled":
                    strip_color = self.strip_program.strip_color()
                    self.strip.set_color(strip_color)

            computation_time = time.time() - computation_start

            if self.screen_program.name != "Disabled":
                time_sum += computation_time
                iteration_count += 1
                if (
                    iteration_count >= adaptive_quality_window
                    or time_sum
                    >= 1.5 * adaptive_quality_window * self.seconds_per_frame
                ):
                    average_computation_time = time_sum / adaptive_quality_window
                    iteration_count = 0
                    time_sum = 0

                    # print(f'average compute ratio: {average_computation_time/self.seconds_per_frame}')
                    if average_computation_time > 0.9 * self.seconds_per_frame:
                        # if the loop takes too long and a screen program is active, it can be reduced in resolution to save time
                        self.screen_program.decrease_resolution()
                    elif average_computation_time < 0.6 * self.seconds_per_frame:
                        # if the loop has time to spare and a screen program is active, we can increase its quality
                        self.screen_program.increase_resolution()

            # print(f'computation took {computation_time:.5f}s')
            try:
                time.sleep(self.seconds_per_frame - computation_time)
            except ValueError:
                pass

    def consumers_changed(self):
        if self.programs["Disabled"].consumers == 3:
            self.loop_active.clear()
        else:
            self.loop_active.set()

    def _set_ring_program(self, program, transient=False):
        # don't allow program change on disconnected devices
        if not self.ring.initialized:
            return

        self.ring_program.release()
        program.use()

        self.last_ring_program = self.ring_program
        self.ring_program = program
        if not transient:
            Setting.objects.filter(key="last_ring_program").update(
                value=self.last_ring_program.name
            )
            Setting.objects.filter(key="ring_program").update(
                value=self.ring_program.name
            )
        self.consumers_changed()

        if program.name == "Disabled":
            self.ring.clear()

    def _set_strip_program(self, program, transient=False):
        # don't allow program change on disconnected devices
        if not self.strip.initialized:
            return

        self.strip_program.release()
        program.use()

        self.last_strip_program = self.strip_program
        self.strip_program = program
        if not transient:
            Setting.objects.filter(key="last_strip_program").update(
                value=self.last_strip_program.name
            )
            Setting.objects.filter(key="strip_program").update(
                value=self.strip_program.name
            )
        self.consumers_changed()

        if program.name == "Disabled":
            self.strip.clear()

    def _set_screen_program(self, program, transient=False):
        if not self.screen.initialized:
            return
        self.screen_program.release()
        program.use()

        self.last_screen_program = self.screen_program
        self.screen_program = program
        if not transient:
            Setting.objects.filter(key="last_screen_program").update(
                value=self.last_screen_program.name
            )
            Setting.objects.filter(key="screen_program").update(
                value=self.screen_program.name
            )
        self.consumers_changed()

    def alarm_started(self):
        # make alarm the current program but don't update the database
        with self.option_lock:
            self.alarm_program.use()
            self.last_fixed_color = self.programs["Fixed"].color
            self._set_ring_program(self.programs["Fixed"], transient=True)
            self._set_strip_program(self.programs["Fixed"], transient=True)
            # the screen program adapts with the alarm and is not changed

    def alarm_stopped(self):
        with self.option_lock:
            self.alarm_program.release()
            self.programs["Fixed"].color = self.last_fixed_color
            self._set_ring_program(self.last_ring_program, transient=True)
            self._set_strip_program(self.last_strip_program, transient=True)

            # read last programs from database, which is still in the state before the alarm
            last_ring_program_name = Setting.objects.get(key="last_ring_program").value
            last_strip_program_name = Setting.objects.get(
                key="last_strip_program"
            ).value
            self.last_ring_program = self.programs[last_ring_program_name]
            self.last_strip_program = self.programs[last_strip_program_name]

    def state_dict(self):
        state_dict = self.base.state_dict()
        state_dict["ring_connected"] = self.ring.initialized
        state_dict["ring_program"] = self.ring_program.name
        state_dict["ring_brightness"] = self.ring.brightness
        state_dict["ring_monochrome"] = self.ring.monochrome
        state_dict["strip_connected"] = self.strip.initialized
        state_dict["strip_program"] = self.strip_program.name
        state_dict["strip_brightness"] = self.strip.brightness
        state_dict["screen_connected"] = self.screen.initialized
        state_dict["screen_program"] = self.screen_program.name
        state_dict["program_speed"] = self.program_speed
        state_dict["fixed_color"] = "#{:02x}{:02x}{:02x}".format(
            *(int(val * 255) for val in self.programs["Fixed"].color)
        )
        return state_dict

    def get_state(self, request):
        state = self.state_dict()
        return JsonResponse(state)

    def update_state(self):
        state_handler.update_state(self.state_dict())

    def index(self, request):
        context = self.base.context(request)
        # programs that have a strip_color or ring_color function are color programs
        # programs that have a draw function are screen programs
        context["color_program_names"] = [
            program.name
            for program in self.programs.values()
            if getattr(program, "ring_colors", None)
        ]
        context["screen_program_names"] = [
            program.name
            for program in self.programs.values()
            if getattr(program, "draw", None)
        ]
        # context['program_names'].remove('Alarm')
        return render(request, "lights.html", context)

    # every option change needs to be synchronized
    # also it changes the state
    def option(func):
        def _decorator(self, request, *args, **kwargs):
            # only privileged users can change options during voting system
            if (
                self.base.settings.voting_system
                and not self.base.user_manager.has_controls(request.user)
            ):
                return HttpResponseForbidden()
            # don't allow option changes during alarm
            if self.base.musiq.player.alarm_playing.is_set():
                return HttpResponseForbidden()
            with self.option_lock:
                try:
                    response = func(self, request, *args, **kwargs)
                    if response is not None:
                        return response
                except (ValueError, IndexError) as e:
                    print("error during lights option: " + str(e))
                    return HttpResponseBadRequest()
                self.update_state()
            return HttpResponse()

        return wraps(func)(_decorator)

    @option
    def set_lights_shortcut(self, request):
        should_enable = request.POST.get("value") == "true"
        is_enabled = (
            self.ring_program.name != "Disabled"
            or self.strip_program.name != "Disabled"
        )
        if should_enable == is_enabled:
            return
        if should_enable:
            self._set_ring_program(self.last_ring_program)
            self._set_strip_program(self.last_strip_program)
        else:
            self._set_ring_program(self.programs["Disabled"])
            self._set_strip_program(self.programs["Disabled"])

    @option
    def set_ring_program(self, request):
        program_name = request.POST.get("program")
        program = self.programs[program_name]
        if program == self.ring_program:
            # the program doesn't change, return immediately
            return
        self._set_ring_program(program)

    @option
    def set_ring_brightness(self, request):
        # raises ValueError on wrong input, caught in option decorator
        value = float(request.POST.get("value"))
        self.ring.brightness = value

    @option
    def set_ring_monochrome(self, request):
        enabled = request.POST.get("value") == "true"
        self.ring.monochrome = enabled

    @option
    def set_strip_program(self, request):
        program_name = request.POST.get("program")
        program = self.programs[program_name]
        if program == self.strip_program:
            # the program doesn't change, return immediately
            return
        self._set_strip_program(program)

    @option
    def set_strip_brightness(self, request):
        # raises ValueError on wrong input, caught in option decorator
        value = float(request.POST.get("value"))
        self.strip.brightness = value

    @option
    def adjust_screen(self, request):
        if self.screen_program.name != "Disabled":
            return HttpResponseBadRequest(
                "Disable the screen program before readjusting"
            )
        self.screen.adjust()

    @option
    def set_screen_program(self, request):
        program_name = request.POST.get("program")
        program = self.programs[program_name]
        if program == self.screen_program:
            # the program doesn't change, return immediately
            return
        self._set_screen_program(program)

    @option
    def set_program_speed(self, request):
        value = float(request.POST.get("value"))
        self.program_speed = value

    @option
    def set_fixed_color(self, request):
        hex_col = request.POST.get("value").lstrip("#")
        # raises IndexError on wrong input, caught in option decorator
        color = tuple(int(hex_col[i : i + 2], 16) / 255 for i in (0, 2, 4))
        self.programs["Fixed"].color = color



from django.db import transaction
from django.shortcuts import render
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.http import JsonResponse
from django.http import HttpResponseBadRequest

from core import models
import core.state_handler as state_handler


class Pad:
    def __init__(self, base):
        self.base = base

    def state_dict(self):
        state_dict = self.base.state_dict()
        state_dict["pad_version"] = models.Pad.objects.get(id=1).version
        return state_dict

    def get_state(self, request):
        state = self.state_dict()
        return JsonResponse(state)

    def update_state(self):
        state_handler.update_state(self.state_dict())

    def index(self, request):
        if not self.base.user_manager.has_pad(request.user):
            raise PermissionDenied
        context = self.base.context(request)

        pad = models.Pad.objects.get_or_create(
            id=1, defaults={"content": "", "version": 0}
        )[0]
        context["pad_version"] = pad.version
        context["pad_content"] = pad.content
        return render(request, "pad.html", context)

    def submit(self, request):
        version = request.POST.get("version")
        content = request.POST.get("content")
        if version is None or version == "" or content is None:
            return HttpResponseBadRequest("No version or content supplied")
        try:
            version = int(version)
        except ValueError:
            return HttpResponseBadRequest("version is not a number")

        version_valid = True
        with transaction.atomic():
            pad = models.Pad.objects.get(id=1)
            current_version = pad.version
            if current_version == version:
                version_valid = True
                pad.version += 1
                pad.content = content
                pad.save()
            else:
                version_valid = False

        if version_valid:
            self.update_state()
            return HttpResponse("Updated Pad")
        else:
            return HttpResponseBadRequest(
                "The content was changed in the meantime, please reload"
            )



from django.conf import settings
from django.http import HttpResponse
from django.http import JsonResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseRedirect
from django.db import transaction
from django.urls import reverse

from core.settings import Settings
from core.musiq.musiq import Musiq
from core.lights.lights import Lights
from core.pad import Pad
from core.user_manager import UserManager
import core.models as models
import core.state_handler as state_handler

import os
import random
import logging


class Base:
    def __init__(self):
        self.logger = logging.getLogger("raveberry")
        self.settings = Settings(self)
        self.user_manager = UserManager(self)
        self.lights = Lights(self)
        self.pad = Pad(self)
        self.musiq = Musiq(self)

    def get_random_hashtag(self):
        if models.Tag.objects.count() == 0:
            return "no hashtags present :("
        index = random.randint(0, models.Tag.objects.count() - 1)
        hashtag = models.Tag.objects.all()[index]
        return hashtag.text

    def increment_counter(self):
        with transaction.atomic():
            counter = models.Counter.objects.get_or_create(id=1, defaults={"value": 0})[
                0
            ]
            counter.value += 1
            counter.save()
        self.update_state()
        return counter.value

    def _get_apk_link(self):
        local_apk = os.path.join(settings.STATIC_ROOT, "apk/shareberry.apk")
        if os.path.isfile(local_apk):
            return os.path.join(settings.STATIC_URL, "apk/shareberry.apk")
        else:
            return "https://github.com/raveberry/shareberry/raw/master/app/release/shareberry.apk"

    def context(self, request):
        self.increment_counter()
        return {
            "voting_system": self.settings.voting_system,
            "hashtag": self.get_random_hashtag(),
            "controls_enabled": self.user_manager.has_controls(request.user),
            "pad_enabled": self.user_manager.has_pad(request.user),
            "is_admin": self.user_manager.is_admin(request.user),
            "apk_link": self._get_apk_link(),
            "spotify_enabled": self.settings.spotify_enabled,
        }

    def state_dict(self):
        # this function constructs a base state dictionary with website wide state
        # pages sending states extend this state dictionary
        return {
            "partymode": self.user_manager.partymode_enabled(),
            "users": self.user_manager.get_count(),
            "visitors": models.Counter.objects.get_or_create(
                id=1, defaults={"value": 0}
            )[0].value,
            "lights_enabled": self.lights.loop_active.is_set(),
            "alarm": self.musiq.player.alarm_playing.is_set(),
            "default_platform": "spotify"
            if self.settings.spotify_enabled
            else "youtube",
        }

    def get_state(self, request):
        state = self.state_dict()
        return JsonResponse(state)

    def update_state(self):
        state_handler.update_state(self.state_dict())

    def submit_hashtag(self, request):
        hashtag = request.POST.get("hashtag")
        if hashtag is None or len(hashtag) == 0:
            return HttpResponseBadRequest()

        if hashtag[0] != "#":
            hashtag = "#" + hashtag
        models.Tag.objects.create(text=hashtag)

        return HttpResponse()

    def logged_in(self, request):
        if request.user.username == "admin":
            return HttpResponseRedirect(reverse("settings"))
        else:
            return HttpResponseRedirect(reverse("musiq"))



from contextlib import contextmanager

from django.conf import settings
from django.db import transaction
from django.db.models import F
from django.http import HttpResponse, HttpResponseBadRequest

import core.musiq.song_utils as song_utils

import youtube_dl
import subprocess
import requests
import pickle
import errno
import time
import json
import os
import threading
import mutagen.easymp4

from urllib.parse import urlparse
from urllib.parse import parse_qs

from core.models import (
    ArchivedSong,
    ArchivedPlaylist,
    PlaylistEntry,
    ArchivedPlaylistQuery,
    RequestLog,
)
from core.musiq.music_provider import SongProvider, PlaylistProvider
from core.util import background_thread


class MyLogger(object):
    def debug(self, msg):
        if settings.DEBUG:
            print(msg)

    def warning(self, msg):
        if settings.DEBUG:
            print(msg)

    def error(self, msg):
        print(msg)


# youtube-dl --format bestaudio[ext=m4a]/best[ext=m4a] --output '%(id)s.%(ext)s --no-playlist --write-thumbnail --default-search auto --add-metadata --embed-thumbnail
def get_ydl_opts():
    return {
        "format": "bestaudio[ext=m4a]/best[ext=m4a]",
        "outtmpl": os.path.join(settings.SONGS_CACHE_DIR, "%(id)s.%(ext)s"),
        "noplaylist": True,
        "no_color": True,
        "writethumbnail": True,
        "default_search": "auto",
        "postprocessors": [
            {"key": "FFmpegMetadata",},
            {
                "key": "EmbedThumbnail",
                # overwrite any thumbnails already present
                "already_have_thumbnail": True,
            },
        ],
        "logger": MyLogger(),
    }


def get_initial_data(html):
    for line in html.split("\n"):
        line = line.strip()
        prefix = 'window["ytInitialData"] = '
        if line.startswith(prefix):
            # strip assignment and semicolon
            initial_data = line[len(prefix) : -1]
            initial_data = json.loads(initial_data)
            return initial_data


# a context that opens a session and loads the cookies file
@contextmanager
def youtube_session():
    session = requests.session()
    try:
        with open(
            os.path.join(settings.BASE_DIR, "config/youtube_cookies.pickle"), "rb"
        ) as f:
            session.cookies.update(pickle.load(f))
    except FileNotFoundError:
        pass

    headers = {
        "User-Agent": youtube_dl.utils.random_user_agent(),
    }
    session.headers.update(headers)
    yield session

    with open(
        os.path.join(settings.BASE_DIR, "config/youtube_cookies.pickle"), "wb"
    ) as f:
        pickle.dump(session.cookies, f)


def get_search_suggestions(query):
    with youtube_session() as session:
        params = {
            "client": "youtube",
            "q": query,
            "xhr": "t",  # this makes the response be a json file
        }
        r = session.get("https://clients1.google.com/complete/search", params=params)
    suggestions = json.loads(r.text)
    # first entry is the query, the second one contains the suggestions
    suggestions = suggestions[1]
    # suggestions are given as tuples; extract the string and skip the query if it occurs identically
    suggestions = [entry[0] for entry in suggestions if entry[0] != query]
    return suggestions


class NoPlaylistException(Exception):
    pass


class Downloader:
    def get_playlist_info(self):
        self.ydl_opts = get_ydl_opts()


class YoutubeSongProvider(SongProvider):
    @staticmethod
    def get_id_from_external_url(url):
        return parse_qs(urlparse(url).query)["v"][0]

    @staticmethod
    def get_id_from_internal_url(url):
        return os.path.splitext(os.path.basename(url[len("file://") :]))[0]

    def __init__(self, musiq, query, key):
        super().__init__(musiq, query, key)
        self.type = "youtube"
        self.info_dict = None
        self.ydl_opts = get_ydl_opts()

    def check_cached(self):
        # TODO: in case query is set but key and id is not, try to extract the id from the query
        # example: https://youtu.be/<id>
        if not self._check_cached():
            return False
        return os.path.isfile(self.get_path())

    def check_downloadable(self):
        try:
            with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
                self.info_dict = ydl.extract_info(self.query, download=False)
        except youtube_dl.utils.DownloadError as e:
            self.error = e
            return False

        # this value is not an exact match, but it's a good approximation
        if "entries" in self.info_dict:
            self.info_dict = self.info_dict["entries"][0]

        self.id = self.info_dict["id"]

        size = self.info_dict["filesize"]
        max_size = self.musiq.base.settings.max_download_size * 1024 * 1024
        if (
            max_size != 0
            and song_utils.path_from_id(self.info_dict["id"]) is None
            and (size is not None and size > max_size)
        ):
            self.error = "Song too long"
            return False
        return True

    @background_thread
    def _download(self, ip, archive, manually_requested):
        error = None
        location = None

        self.placeholder = {"query": self.query, "replaced_by": None}
        self.musiq.placeholders.append(self.placeholder)
        self.musiq.update_state()

        try:
            with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
                ydl.download([self.get_external_url()])

            location = self.get_path()
            base = os.path.splitext(location)[0]
            thumbnail = base + ".jpg"
            try:
                os.remove(thumbnail)
            except FileNotFoundError:
                self.musiq.base.logger.info(
                    "tried to delete " + thumbnail + " but does not exist"
                )

        except youtube_dl.utils.DownloadError as e:
            error = e

        if error is not None or location is None:
            self.musiq.logger.error(
                "accessible video could not be downloaded: " + str(self.id)
            )
            self.musiq.logger.error(error)
            self.musiq.logger.error("location: " + str(location))
            self.musiq.placeholders.remove(self.placeholder)
            self.musiq.update_state()
            return
        self.enqueue(ip, archive=archive, manually_requested=manually_requested)

    def download(self, ip, background=True, archive=True, manually_requested=True):
        # check if file was already downloaded and only download if necessary
        if os.path.isfile(self.get_path()):
            self.enqueue(ip, archive=archive, manually_requested=manually_requested)
        else:
            thread = self._download(ip, archive, manually_requested)
            if not background:
                thread.join()
        return True

    def get_metadata(self):
        metadata = song_utils.get_metadata(self.get_path())

        metadata["internal_url"] = self.get_internal_url()
        metadata["external_url"] = "https://www.youtube.com/watch?v=" + self.id
        if not metadata["title"]:
            metadata["title"] = metadata["external_url"]

        return metadata

    def get_path(self):
        path = os.path.join(settings.SONGS_CACHE_DIR, self.id + ".m4a")
        path = path.replace("~", os.environ["HOME"])
        path = os.path.abspath(path)
        return path

    def get_internal_url(self):
        return "file://" + self.get_path()

    def get_external_url(self):
        return "https://www.youtube.com/watch?v=" + self.id

    def get_suggestion(self):
        with youtube_session() as session:
            r = session.get(self.get_external_url())

        initial_data = get_initial_data(r.text)
        # fmt: off
        url = initial_data["contents"]["twoColumnWatchNextResults"]["secondaryResults"] \
            ["secondaryResults"]["results"][0]["compactAutoplayRenderer"]["contents"][0] \
            ["compactVideoRenderer"]["navigationEndpoint"]["commandMetadata"] \
            ["webCommandMetadata"]["url"]
        # fmt: on
        return "https://www.youtube.com" + url

    def request_radio(self, ip):
        radio_id = "RD" + self.id

        provider = YoutubePlaylistProvider(self.musiq, "radio for " + self.id, None)
        provider.id = radio_id
        if not provider.download(ip):
            return HttpResponseBadRequest(provider.error)
        return HttpResponse("queueing radio")


class YoutubePlaylistProvider(PlaylistProvider):
    @staticmethod
    def get_id_from_external_url(url):
        try:
            list_id = parse_qs(urlparse(url).query)["list"][0]
        except KeyError:
            return None
        return list_id

    def __init__(self, musiq, query, key):
        super().__init__(musiq, query, key)
        self.type = "youtube"
        self.ydl_opts = get_ydl_opts()
        del self.ydl_opts["noplaylist"]
        self.ydl_opts["extract_flat"] = True

    def is_radio(self):
        return self.id.startswith("RD")

    def search_id(self):
        with youtube_session() as session:
            params = {
                "search_query": self.query,
                # this is the value that youtube uses to filter for playlists only
                "sp": "EgQQA1AD",
            }
            r = session.get("https://www.youtube.com/results", params=params)

        initial_data = get_initial_data(r.text)
        # fmt: off
        section_renderers = initial_data["contents"]["twoColumnSearchResultsRenderer"]\
            ["primaryContents"]["sectionListRenderer"]["contents"]
        # fmt: on

        list_id = None
        for section_renderer in section_renderers:
            search_results = section_renderer["itemSectionRenderer"]["contents"]

            try:
                list_id = next(
                    res["playlistRenderer"]["playlistId"]
                    for res in search_results
                    if "playlistRenderer" in res
                )
                break
            except StopIteration:
                # the search result did not contain the list id
                pass

        return list_id

    def fetch_metadata(self):
        # in case of a radio playist, restrict the number of songs that are downloaded
        if self.is_radio():
            self.ydl_opts["playlistend"] = self.musiq.base.settings.max_playlist_items

        with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
            info_dict = ydl.extract_info(self.id, download=False)

        if info_dict["_type"] != "playlist" or "entries" not in info_dict:
            # query was not a playlist url -> search for the query
            assert False

        assert self.id == info_dict["id"]
        if "title" in info_dict:
            self.title = info_dict["title"]
        for entry in info_dict["entries"]:
            self.urls.append("https://www.youtube.com/watch?v=" + entry["id"])
        assert self.key is None



from django.shortcuts import render
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseForbidden
from django.http import JsonResponse
from django.utils import dateparse
from django.utils import timezone
from django.db import models
from django.conf import settings
from mutagen import MutagenError

from core.util import background_thread
from main import settings
from core.models import Setting, ArchivedSong, ArchivedPlaylist, PlaylistEntry
from core.models import PlayLog
from core.models import RequestLog
import core.state_handler as state_handler
import core.musiq.song_utils as song_utils

from functools import wraps
from datetime import timedelta
from dateutil import tz
import urllib.request
import configparser
import subprocess
import threading
import shutil
import time
import math
import os
import re


class Settings:
    def __init__(self, base):
        self.base = base
        self.voting_system = self.get_setting("voting_system", "False") == "True"
        self.logging_enabled = self.get_setting("logging_enabled", "True") == "True"
        self.people_to_party = int(self.get_setting("people_to_party", 3))
        self.alarm_probability = float(self.get_setting("alarm_probability", 0))
        self.downvotes_to_kick = int(self.get_setting("downvotes_to_kick", 3))
        self.max_download_size = int(self.get_setting("max_download_size", 10))
        self.max_playlist_items = int(self.get_setting("max_playlist_items", 10))
        self.spotify_username = self.get_setting("spotify_username", "")
        self.spotify_password = self.get_setting("spotify_password", "")
        self.spotify_client_id = self.get_setting("spotify_client_id", "")
        self.spotify_client_secret = self.get_setting("spotify_client_secret", "")

        self.spotify_enabled = False
        self._check_spotify()
        self._check_internet()
        self.bluetoothctl = None
        self.bluetooth_devices = []
        self.homewifi = self.get_setting("homewifi", "")
        self.scan_progress = "0 / 0 / 0"

    def get_setting(self, key, default):
        return Setting.objects.get_or_create(key=key, defaults={"value": default})[
            0
        ].value

    def state_dict(self):
        state_dict = self.base.state_dict()
        state_dict["voting_system"] = self.voting_system
        state_dict["logging_enabled"] = self.logging_enabled
        state_dict["people_to_party"] = self.people_to_party
        state_dict["alarm_probability"] = self.alarm_probability
        state_dict["downvotes_to_kick"] = self.downvotes_to_kick
        state_dict["max_download_size"] = self.max_download_size
        state_dict["max_playlist_items"] = self.max_playlist_items
        state_dict["has_internet"] = self.has_internet

        state_dict["spotify_credentials_valid"] = self.spotify_enabled

        state_dict["bluetooth_scanning"] = self.bluetoothctl is not None
        state_dict["bluetooth_devices"] = self.bluetooth_devices

        try:
            with open(os.path.join(settings.BASE_DIR, "config/homewifi")) as f:
                state_dict["homewifi_ssid"] = f.read()
        except FileNotFoundError:
            state_dict["homewifi_ssid"] = ""

        state_dict["scan_progress"] = self.scan_progress

        try:
            state_dict["homewifi_enabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/homewifi_enabled"]) != 0
            )
            state_dict["events_enabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/events_enabled"]) != 0
            )
            state_dict["hotspot_enabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/hotspot_enabled"]) != 0
            )
            state_dict["wifi_protected"] = (
                subprocess.call(["/usr/local/sbin/raveberry/wifi_protected"]) != 0
            )
            state_dict["tunneling_enabled"] = (
                subprocess.call(["sudo", "/usr/local/sbin/raveberry/tunneling_enabled"])
                != 0
            )
            state_dict["remote_enabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/remote_enabled"]) != 0
            )
        except FileNotFoundError:
            self.base.logger.error("scripts not installed")

        return state_dict

    def get_state(self, request):
        state = self.state_dict()
        return JsonResponse(state)

    def update_state(self):
        state_handler.update_state(self.state_dict())

    def index(self, request):
        if not self.base.user_manager.is_admin(request.user):
            raise PermissionDenied
        context = self.base.context(request)
        return render(request, "settings.html", context)

    def _check_spotify(self, credentials_changed=False):
        if not self.spotify_client_id or not self.spotify_client_secret:
            self.spotify_enabled = False
            return HttpResponseBadRequest("Not all credentials provided")
        if subprocess.run(
            ["systemctl", "is-active", "mopidy"], stdout=subprocess.DEVNULL
        ).returncode:
            return self._check_spotify_user(credentials_changed=credentials_changed)
        else:
            return self._check_spotify_service(credentials_changed=credentials_changed)

    def _check_spotify_user(self, credentials_changed=False):
        self.spotify_enabled = False
        # TODO use credentials from settings
        config = subprocess.run(
            ["mopidy", "config"], stdout=subprocess.PIPE, universal_newlines=True
        ).stdout
        parser = configparser.ConfigParser()
        parser.read_string(config)
        try:
            if parser["spotify"]["enabled"] == "true":
                self.spotify_enabled = True
                return HttpResponse("Login probably successful")
        except KeyError:
            pass
        return HttpResponseBadRequest("Config is invalid")

    def _check_spotify_service(self, credentials_changed=False):
        if credentials_changed:
            if shutil.which("cava"):
                # if cava is installed, use the visualization config for mopidy
                config_file = os.path.join(settings.BASE_DIR, "setup/mopidy_cava.conf")
            else:
                config_file = os.path.join(settings.BASE_DIR, "setup/mopidy.conf")

            subprocess.call(
                [
                    "sudo",
                    "/usr/local/sbin/raveberry/update_mopidy_config",
                    config_file,
                    self.spotify_username,
                    self.spotify_password,
                    self.spotify_client_id,
                    self.spotify_client_secret,
                ]
            )
            subprocess.call(["sudo", "/usr/local/sbin/raveberry/restart_mopidy"])

            # wait for mopidy to try spotify login
            time.sleep(5)

        # check the mopidy log and see if there is a spotify login error since the last restart
        log = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/read_mopidy_log"],
            universal_newlines=True,
        )
        login_error = False
        response = None
        for line in log.split("\n")[::-1]:
            if line.startswith("ERROR") and "spotify.session" in line:
                login_error = True
                response = HttpResponseBadRequest("User or Password are wrong")
                break
            elif line.startswith("ERROR") and "mopidy_spotify.web" in line:
                login_error = True
                response = HttpResponseBadRequest(
                    "Client ID or Client Secret are wrong or expired"
                )
                break
            if (
                line.startswith("WARNING")
                and "The extension has been automatically disabled" in line
            ):
                login_error = True
                response = HttpResponseBadRequest("Configuration Error")
                break
            elif line.startswith("Started Mopidy music server."):
                response = HttpResponse("Login successful")
                break
        else:
            # there were too many lines in the log, could not determine whether there was an error
            # leave spotify_enabled status as it is
            return HttpResponseBadRequest("Could not check credentials")

        if not login_error:
            self.spotify_enabled = True
        else:
            self.spotify_enabled = False
        return response

    def _check_internet(self):
        response = subprocess.call(
            ["ping", "-c", "1", "-W", "3", "1.1.1.1"], stdout=subprocess.DEVNULL
        )
        if response == 0:
            self.has_internet = True
        else:
            self.has_internet = False

    # settings can only be changed by admin
    def option(func):
        def _decorator(self, request, *args, **kwargs):
            # don't allow option changes during alarm
            if request.user.username != "admin":
                return HttpResponseForbidden()
            response = func(self, request, *args, **kwargs)
            self.update_state()
            if response is not None:
                return response
            return HttpResponse()

        return wraps(func)(_decorator)

    @option
    def set_voting_system(self, request):
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="voting_system").update(value=enabled)
        self.voting_system = enabled

    @option
    def set_logging_enabled(self, request):
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="logging_enabled").update(value=enabled)
        self.logging_enabled = enabled

    @option
    def set_people_to_party(self, request):
        value = int(request.POST.get("value"))
        Setting.objects.filter(key="people_to_party").update(value=value)
        self.people_to_party = value

    @option
    def set_alarm_probability(self, request):
        value = float(request.POST.get("value"))
        Setting.objects.filter(key="alarm_probability").update(value=value)
        self.alarm_probability = value

    @option
    def set_downvotes_to_kick(self, request):
        value = int(request.POST.get("value"))
        Setting.objects.filter(key="downvotes_to_kick").update(value=value)
        self.downvotes_to_kick = value

    @option
    def set_max_download_size(self, request):
        value = int(request.POST.get("value"))
        Setting.objects.filter(key="max_download_size").update(value=value)
        self.max_download_size = value

    @option
    def set_max_playlist_items(self, request):
        value = int(request.POST.get("value"))
        Setting.objects.filter(key="max_playlist_items").update(value=value)
        self.max_playlist_items = value

    @option
    def check_internet(self, request):
        self._check_internet()

    @option
    def update_user_count(self, request):
        self.base.user_manager.update_user_count()

    @option
    def check_spotify_credentials(self, request):
        return self._check_spotify()

    @option
    def set_spotify_credentials(self, request):
        username = request.POST.get("username")
        password = request.POST.get("password")
        client_id = request.POST.get("client_id")
        client_secret = request.POST.get("client_secret")

        if (
            username is None
            or username == ""
            or password is None
            or password == ""
            or client_id is None
            or client_id == ""
            or client_secret is None
            or client_secret == ""
        ):
            return HttpResponseBadRequest("All fields are required")

        self.spotify_username = username
        self.spotify_password = password
        self.spotify_client_id = client_id
        self.spotify_client_secret = client_secret

        Setting.objects.filter(key="spotify_username").update(
            value=self.spotify_username
        )
        Setting.objects.filter(key="spotify_password").update(
            value=self.spotify_password
        )
        Setting.objects.filter(key="spotify_client_id").update(
            value=self.spotify_client_id
        )
        Setting.objects.filter(key="spotify_client_secret").update(
            value=self.spotify_client_secret
        )

        return self._check_spotify(credentials_changed=True)

    def _get_bluetoothctl_line(self):
        # Note: this variable is not guarded by a lock.
        # But there should only be one admin accessing these bluetooth functions anyway.
        if self.bluetoothctl is None:
            return ""
        line = self.bluetoothctl.stdout.readline().decode()
        ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
        line = ansi_escape.sub("", line)
        line = line.strip()
        return line

    def _stop_bluetoothctl(self):
        self.bluetoothctl.stdin.close()
        self.bluetoothctl.wait()
        self.bluetoothctl = None

    @option
    def set_bluetooth_scanning(self, request):
        enabled = request.POST.get("value") == "true"
        if enabled:
            if self.bluetoothctl is not None:
                return HttpResponseBadRequest("Already Scanning")
            self.bluetooth_devices = []
            self.bluetoothctl = subprocess.Popen(
                ["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
            )

            self.bluetoothctl.stdin.write(b"devices\n")
            self.bluetoothctl.stdin.write(b"scan on\n")
            self.bluetoothctl.stdin.flush()
            while True:
                line = self._get_bluetoothctl_line()
                if not line:
                    break
                # match old devices
                match = re.match("Device (\S*) (.*)", line)
                # match newly scanned devices
                # We need the '.*' at the beginning of the line to account for control sequences
                if not match:
                    match = re.match(".*\[NEW\] Device (\S*) (.*)", line)
                if match:
                    address = match.group(1)
                    name = match.group(2)
                    # filter unnamed devices
                    # devices named after their address are no speakers
                    if re.match("[A-Z0-9][A-Z0-9](-[A-Z0-9][A-Z0-9]){5}", name):
                        continue
                    self.bluetooth_devices.append(
                        {"address": address, "name": name,}
                    )
                    self.update_state()
        else:
            if self.bluetoothctl is None:
                return HttpResponseBadRequest("Currently not scanning")
            self._stop_bluetoothctl()

    @option
    def connect_bluetooth(self, request):
        address = request.POST.get("address")
        if self.bluetoothctl is not None:
            return HttpResponseBadRequest("Stop scanning before connecting")
        if address is None or address == "":
            return HttpResponseBadRequest("No device selected")

        self.bluetoothctl = subprocess.Popen(
            ["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        error = ""

        # A Function that acts as a timeout for unexpected errors (or timeouts)
        @background_thread
        def timeout():
            time.sleep(20)
            error = "Timed out"
            if self.bluetoothctl is not None:
                self._stop_bluetoothctl()

        self.bluetoothctl.stdin.write(b"pair " + address.encode() + b"\n")
        self.bluetoothctl.stdin.flush()
        while True:
            line = self._get_bluetoothctl_line()
            if not line:
                break
            if re.match(".*Device " + address + " not available", line):
                error = "Device unavailable"
                break
            elif re.match(".*Failed to pair: org.bluez.Error.AlreadyExists", line):
                break
            elif re.match(".*Pairing successful", line):
                break

        if error:
            self._stop_bluetoothctl()
            return HttpResponseBadRequest(error)

        self.bluetoothctl.stdin.write(b"connect " + address.encode() + b"\n")
        self.bluetoothctl.stdin.flush()
        while True:
            line = self._get_bluetoothctl_line()
            if not line:
                break
            if re.match(".*Device " + address + " not available", line):
                error = "Device unavailable"
                break
            elif re.match(".*Failed to connect: org.bluez.Error.Failed", line):
                error = "Connect Failed"
                break
            elif re.match(".*Failed to connect: org.bluez.Error.InProgress", line):
                error = "Connect in progress"
                break
            elif re.match(".*Connection successful", line):
                break
        # trust the device to automatically reconnect when it is available again
        self.bluetoothctl.stdin.write(b"trust " + address.encode() + b"\n")
        self.bluetoothctl.stdin.flush()

        self._stop_bluetoothctl()
        if error:
            return HttpResponseBadRequest(error)

        # parse the sink number of the bluetooth device from pactl
        sinks = subprocess.check_output(
            "pactl list short sinks".split(), universal_newlines=True
        )
        bluetooth_sink = "2"
        for sink in sinks.split("\n"):
            if "bluez" in sink:
                bluetooth_sink = sink[0]
                break
        subprocess.call(
            f"pactl set-default-sink {bluetooth_sink}".split(),
            stdout=subprocess.DEVNULL,
        )
        # restart mopidy to apply audio device change
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/restart_mopidy"])

        return HttpResponse("Connected")

    @option
    def disconnect_bluetooth(self, request):
        subprocess.call("pactl set-default-sink 0".split(), stdout=subprocess.DEVNULL)
        # restart mopidy to apply audio device change
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/restart_mopidy"])
        return HttpResponse("Disconnected")

    @option
    def available_ssids(self, request):
        output = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/list_available_ssids"]
        )
        output = output.decode()
        ssids = output.split("\n")
        return JsonResponse(ssids[:-1], safe=False)

    @option
    def connect_to_wifi(self, request):
        ssid = request.POST.get("ssid")
        password = request.POST.get("password")
        if ssid is None or password is None or ssid == "" or password == "":
            return HttpResponseBadRequest("Please provide both SSID and password")
        try:
            output = subprocess.check_output(
                ["sudo", "/usr/local/sbin/raveberry/connect_to_wifi", ssid, password]
            )
            output = output.decode()
            return HttpResponse(output)
        except subprocess.CalledProcessError as e:
            output = e.output.decode()
            return HttpResponseBadRequest(output)

    @option
    def disable_homewifi(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_homewifi"])

    @option
    def enable_homewifi(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_homewifi"])

    @option
    def stored_ssids(self, request):
        output = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/list_stored_ssids"]
        )
        output = output.decode()
        ssids = output.split("\n")
        return JsonResponse(ssids[:-1], safe=False)

    @option
    def set_homewifi_ssid(self, request):
        homewifi_ssid = request.POST.get("homewifi_ssid")
        with open(os.path.join(settings.BASE_DIR, "config/homewifi"), "w+") as f:
            f.write(homewifi_ssid)

    @option
    def list_subdirectories(self, request):
        path = request.GET.get("path")
        basedir, subdirpart = os.path.split(path)
        if path == "":
            suggestions = ["/"]
        elif os.path.isdir(basedir):
            suggestions = [
                os.path.join(basedir, subdir + "/")
                for subdir in next(os.walk(basedir))[1]
                if subdir.lower().startswith(subdirpart.lower())
            ]
            suggestions.sort()
        else:
            suggestions = ["not a valid directory"]
        if not suggestions:
            suggestions = ["not a valid directory"]
        return JsonResponse(suggestions, safe=False)

    @option
    def scan_library(self, request):
        library_path = request.POST.get("library_path")

        if not os.path.isdir(library_path):
            return HttpResponseBadRequest("not a directory")
        library_path = os.path.abspath(library_path)

        self.scan_progress = "0 / 0 / 0"
        self.update_state()

        self._scan_library(library_path)

        return HttpResponse(
            f"started scanning in {library_path}. This could take a while"
        )

    @background_thread
    def _scan_library(self, library_path):
        scan_start = time.time()
        last_update = scan_start
        update_frequency = 0.5
        filecount = 0
        for (dirpath, _, filenames) in os.walk(library_path):
            now = time.time()
            if now - last_update > update_frequency:
                last_update = now
                self.scan_progress = f"{filecount} / 0 / 0"
                self.update_state()
            if os.path.abspath(dirpath) == os.path.abspath(settings.SONGS_CACHE_DIR):
                # do not add files handled by raveberry as local files
                continue
            filecount += len(filenames)

        library_link = os.path.join(settings.SONGS_CACHE_DIR, "local_library")
        try:
            os.remove(library_link)
        except FileNotFoundError:
            pass
        os.symlink(library_path, library_link)

        self.base.logger.info(f"started scanning in {library_path}")

        self.scan_progress = f"{filecount} / 0 / 0"
        self.update_state()

        files_scanned = 0
        files_added = 0
        for (dirpath, _, filenames) in os.walk(library_path):
            if os.path.abspath(dirpath) == os.path.abspath(settings.SONGS_CACHE_DIR):
                # do not add files handled by raveberry as local files
                continue
            now = time.time()
            if now - last_update > update_frequency:
                last_update = now
                self.scan_progress = f"{filecount} / {files_scanned} / {files_added}"
                self.update_state()
            for filename in filenames:
                files_scanned += 1
                path = os.path.join(dirpath, filename)
                try:
                    metadata = song_utils.get_metadata(path)
                except (ValueError, MutagenError):
                    # the given file could not be parsed and will not be added to the database
                    pass
                else:
                    library_relative_path = path[len(library_path) + 1 :]
                    external_url = os.path.join("local_library", library_relative_path)
                    if not ArchivedSong.objects.filter(url=external_url).exists():
                        files_added += 1
                        ArchivedSong.objects.create(
                            url=external_url,
                            artist=metadata["artist"],
                            title=metadata["title"],
                            counter=0,
                        )

        assert files_scanned == filecount
        self.scan_progress = f"{filecount} / {files_scanned} / {files_added}"
        self.update_state()

        self.base.logger.info(f"done scanning in {library_path}")

    @option
    def create_playlists(self, request):
        library_link = os.path.join(settings.SONGS_CACHE_DIR, "local_library")
        if not os.path.islink(library_link):
            return HttpResponseBadRequest("No library set")

        self.scan_progress = f"0 / 0 / 0"
        self.update_state()

        self._create_playlists()

        return HttpResponse(f"started creating playlsts. This could take a while")

    @background_thread
    def _create_playlists(self):
        local_files = ArchivedSong.objects.filter(
            url__startswith="local_library"
        ).count()

        library_link = os.path.join(settings.SONGS_CACHE_DIR, "local_library")
        library_path = os.path.abspath(library_link)

        self.base.logger.info(f"started creating playlists in {library_path}")

        self.scan_progress = f"{local_files} / 0 / 0"
        self.update_state()

        scan_start = time.time()
        last_update = scan_start
        update_frequency = 0.5
        files_processed = 0
        files_added = 0
        for (dirpath, _, filenames) in os.walk(library_path):
            now = time.time()
            if now - last_update > update_frequency:
                last_update = now
                self.scan_progress = (
                    f"{local_files} / {files_processed} / {files_added}"
                )
                self.update_state()

            song_urls = []
            # unfortunately there is no way to access track numbers accross different file types
            # so we have to add songs to playlists alphabetically
            for filename in sorted(filenames):
                path = os.path.join(dirpath, filename)
                library_relative_path = path[len(library_path) + 1 :]
                external_url = os.path.join("local_library", library_relative_path)
                if ArchivedSong.objects.filter(url=external_url).exists():
                    files_processed += 1
                    song_urls.append(external_url)

            if not song_urls:
                continue

            playlist_id = os.path.join(
                "local_library", dirpath[len(library_path) + 1 :]
            )
            playlist_title = os.path.split(dirpath)[1]
            playlist, created = ArchivedPlaylist.objects.get_or_create(
                list_id=playlist_id, title=playlist_title, counter=0
            )
            if not created:
                # this playlist already exists, skip
                continue

            song_index = 0
            for external_url in song_urls:
                PlaylistEntry.objects.create(
                    playlist=playlist, index=song_index, url=external_url,
                )
                files_added += 1
                song_index += 1

        self.scan_progress = f"{local_files} / {files_processed} / {files_added}"
        self.update_state()

        self.base.logger.info(f"finished creating playlists in {library_path}")

    @option
    def analyse(self, request):
        startdate = request.POST.get("startdate")
        starttime = request.POST.get("starttime")
        enddate = request.POST.get("enddate")
        endtime = request.POST.get("endtime")
        if (
            startdate is None
            or startdate == ""
            or starttime is None
            or starttime == ""
            or enddate is None
            or enddate == ""
            or endtime is None
            or endtime == ""
        ):
            return HttpResponseBadRequest("All fields are required")

        start = dateparse.parse_datetime(startdate + "T" + starttime)
        end = dateparse.parse_datetime(enddate + "T" + endtime)

        if start is None or end is None:
            return HttpResponseBadRequest("invalid start-/endtime given")
        if start >= end:
            return HttpResponseBadRequest("start has to be before end")

        start = timezone.make_aware(start)
        end = timezone.make_aware(end)

        played = (
            PlayLog.objects.all().filter(created__gte=start).filter(created__lt=end)
        )
        requested = (
            RequestLog.objects.all().filter(created__gte=start).filter(created__lt=end)
        )
        played_count = (
            played.values("song__url", "song__artist", "song__title")
            .values(
                "song__url",
                "song__artist",
                "song__title",
                count=models.Count("song__url"),
            )
            .order_by("-count")
        )
        played_votes = (
            PlayLog.objects.all()
            .filter(created__gte=start)
            .filter(created__lt=end)
            .order_by("-votes")
        )
        devices = requested.values("address").values(
            "address", count=models.Count("address")
        )

        response = {}
        response["songs_played"] = len(played)
        response["most_played_song"] = (
            song_utils.displayname(
                played_count[0]["song__artist"], played_count[0]["song__title"]
            )
            + " ("
            + str(played_count[0]["count"])
            + ")"
        )
        response["highest_voted_song"] = (
            played_votes[0].song.displayname() + " (" + str(played_votes[0].votes) + ")"
        )
        response["most_active_device"] = (
            devices[0]["address"] + " (" + str(devices[0]["count"]) + ")"
        )
        requested_by_ip = requested.filter(address=devices[0]["address"])
        for i in range(6):
            if i >= len(requested_by_ip):
                break
            response["most_active_device"] += "\n"
            if i == 5:
                response["most_active_device"] += "..."
            else:
                response["most_active_device"] += requested_by_ip[i].song.displayname()

        binsize = 3600
        number_of_bins = math.ceil((end - start).total_seconds() / binsize)
        request_bins = [0 for _ in range(number_of_bins)]

        for r in requested:
            seconds = (r.created - start).total_seconds()
            index = int(seconds / binsize)
            request_bins[index] += 1

        current_time = start
        current_index = 0
        response["request_activity"] = ""
        while current_time < end:
            response["request_activity"] += current_time.strftime("%H:%M")
            response["request_activity"] += ":\t" + str(request_bins[current_index])
            response["request_activity"] += "\n"
            current_time += timedelta(seconds=binsize)
            current_index += 1

        localtz = tz.gettz(settings.TIME_ZONE)
        playlist = ""
        for log in played:
            localtime = log.created.astimezone(localtz)
            playlist += "[{:02d}:{:02d}] {}\n".format(
                localtime.hour, localtime.minute, log.song.displayname()
            )
        response["playlist"] = playlist

        return JsonResponse(response)

    @option
    def disable_events(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_events"])

    @option
    def enable_events(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_events"])

    @option
    def disable_hotspot(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_hotspot"])

    @option
    def enable_hotspot(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_hotspot"])

    @option
    def unprotect_wifi(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/unprotect_wifi"])

    @option
    def protect_wifi(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/protect_wifi"])

    @option
    def disable_tunneling(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_tunneling"])

    @option
    def enable_tunneling(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_tunneling"])

    @option
    def disable_remote(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_remote"])

    @option
    def enable_remote(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_remote"])

    @option
    def reboot_server(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/reboot_server"])

    @option
    def reboot_system(self, request):
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/reboot_system"])



from django.conf import settings

from urllib.parse import urlparse
from urllib.parse import parse_qs
import os
import mutagen.easymp4


def path_from_id(song_id):
    path = os.path.join(settings.SONGS_CACHE_DIR, song_id + ".m4a")
    path = path.replace("~", os.environ["HOME"])
    path = os.path.abspath(path)
    if os.path.isfile(path):
        return path
    else:
        return None


def path_from_url(url):
    """ returns the file path for the song corresponding to the given url or None if the url is not cached """
    song_id = id_from_url(url)
    return path_from_id(song_id)


def id_from_url(url):
    return parse_qs(urlparse(url).query)["v"][0]


def is_radio(id_or_url):
    return id_or_url.split("&list=")[-1].startswith("RD")


def determine_playlist_type(archived_playlist):
    # use the url of the first song in the playlist to determine the platform where the playlist is from
    first_song_url = archived_playlist.entries.first().url
    if first_song_url.startswith("local_library/"):
        return "local"
    elif first_song_url.startswith("https://www.youtube.com/"):
        return "youtube"
    elif first_song_url.startswith("https://open.spotify.com/"):
        return "spotify"
    else:
        return None


def format_seconds(seconds):
    hours, seconds = seconds // 3600, seconds % 3600
    minutes, seconds = seconds // 60, seconds % 60

    formatted = ""
    if hours > 0:
        formatted += "{:02d}:".format(int(hours))
    formatted += "{0:02d}:{1:02d}".format(int(minutes), int(seconds))
    return formatted


def get_duration(location):
    parsed = mutagen.File(location)
    return parsed.info.length


def gather_metadata(location):
    pass


def decide_deletion(location):
    return
    """ takes the path to a song file and deletes it if it is larger than a threshold """
    """size = os.path.getsize(location)
    if size > settings.MAX_SONG_SIZE:
        os.remove(location)"""


def displayname(artist, title):
    if artist == "":
        return title
    else:
        return artist + " – " + title


def format_seconds(seconds):
    hours, seconds = seconds // 3600, seconds % 3600
    minutes, seconds = seconds // 60, seconds % 60

    formatted = ""
    if hours > 0:
        formatted += "{:02d}:".format(int(hours))
    formatted += "{0:02d}:{1:02d}".format(int(minutes), int(seconds))
    return formatted


def get_metadata(path):
    """gathers the metadata for the song at the given location.
    'title' and 'duration' is read from tags, the 'url' is built from the location"""

    parsed = mutagen.File(path, easy=True)
    if parsed is None:
        raise ValueError
    metadata = dict()

    if parsed.tags is not None:
        if "artist" in parsed.tags:
            metadata["artist"] = parsed.tags["artist"][0]
        if "title" in parsed.tags:
            metadata["title"] = parsed.tags["title"][0]
    if "artist" not in metadata:
        metadata["artist"] = ""
    if "title" not in metadata:
        metadata["title"] = os.path.split(path)[1]
    if parsed.info is not None and parsed.info.length is not None:
        metadata["duration"] = parsed.info.length
    else:
        metadata["duration"] = -1

    return metadata



from django.conf import settings
from django.shortcuts import render
from django.http import HttpResponse
from django.http import JsonResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseServerError
from django.core import serializers
from django.forms.models import model_to_dict
from django.views.decorators.csrf import csrf_exempt

from core.models import QueuedSong
from core.models import CurrentSong
from core.models import ArchivedSong
from core.musiq.localdrive import LocalSongProvider
from core.musiq.music_provider import SongProvider, PlaylistProvider
from core.musiq.suggestions import Suggestions
from core.musiq.player import Player
from core.musiq.song_queue import SongQueue
from core.musiq.youtube import (
    YoutubeSongProvider,
    NoPlaylistException,
    YoutubePlaylistProvider,
)
from core.musiq.spotify import SpotifySongProvider, SpotifyPlaylistProvider
import core.musiq.song_utils as song_utils
import core.state_handler as state_handler

import youtube_dl
import re
import threading

import time
import logging
import ipware


class Musiq:
    def __init__(self, base):
        self.base = base

        self.logger = logging.getLogger("raveberry")

        self.suggestions = Suggestions(self)

        self.queue = QueuedSong.objects
        self.placeholders = []

        self.player = Player(self)
        self.player.start()

    def _request_music(
        self, ip, query, key, playlist, platform, archive=True, manually_requested=True
    ):
        providers = []

        if playlist:
            if key is not None:
                # an archived song was requested. The key determines the SongProvider (Youtube or Spotify)
                provider = PlaylistProvider.create(self, query, key)
                if provider is None:
                    return HttpResponseBadRequest(
                        "No provider found for requested playlist"
                    )
                providers.append(provider)
            else:
                # try to use spotify if the user did not specifically request youtube
                if platform is None or platform == "spotify":
                    if self.base.settings.spotify_enabled:
                        providers.append(SpotifyPlaylistProvider(self, query, key))
                # use Youtube as a fallback
                providers.append(YoutubePlaylistProvider(self, query, key))
        else:
            if key is not None:
                # an archived song was requested. The key determines the SongProvider (Youtube or Spotify)
                provider = SongProvider.create(self, query, key)
                if provider is None:
                    return HttpResponseBadRequest(
                        "No provider found for requested song"
                    )
                providers.append(provider)
            else:
                if platform == "local":
                    # if a local provider was requested, use only this one as it can only come from the database -> it will probably exist
                    providers.append(LocalSongProvider(self, query, key))
                else:
                    # try to use spotify if the user did not specifically request youtube
                    if platform is None or platform == "spotify":
                        if self.base.settings.spotify_enabled:
                            providers.append(SpotifySongProvider(self, query, key))
                    # use Youtube as a fallback
                    providers.append(YoutubeSongProvider(self, query, key))

        fallback = False
        used_provider = None
        for i, provider in enumerate(providers):
            if not provider.check_cached():
                if not provider.check_downloadable():
                    # this provider cannot provide this song, use the next provider
                    # if this was the last provider, show its error
                    if i == len(providers) - 1:
                        return HttpResponseBadRequest(provider.error)
                    fallback = True
                    continue
                if not provider.download(
                    ip, archive=archive, manually_requested=manually_requested
                ):
                    return HttpResponseBadRequest(provider.error)
            else:
                provider.enqueue(
                    ip, archive=archive, manually_requested=manually_requested
                )
            # the current provider could provide the song, don't try the other ones
            used_provider = provider
            break
        message = used_provider.ok_message
        if fallback:
            message = message + " (used fallback)"
        return HttpResponse(message)

    def request_music(self, request):
        key = request.POST.get("key")
        playlist = request.POST.get("playlist") == "true"
        query = request.POST.get("query")
        platform = request.POST.get("platform")

        # only get ip on user requests
        if self.base.settings.logging_enabled:
            ip, is_routable = ipware.get_client_ip(request)
            if ip is None:
                ip = ""
        else:
            ip = ""

        return self._request_music(ip, query, key, playlist, platform)

    def request_radio(self, request):
        # only get ip on user requests
        if self.base.settings.logging_enabled:
            ip, is_routable = ipware.get_client_ip(request)
            if ip is None:
                ip = ""
        else:
            ip = ""

        try:
            current_song = CurrentSong.objects.get()
        except CurrentSong.DoesNotExist:
            return HttpResponseBadRequest("Need a song to play the radio")
        provider = SongProvider.create(self, external_url=current_song.external_url)
        return provider.request_radio(ip)

    @csrf_exempt
    def post_song(self, request):
        return self.request_music(request)

    def index(self, request):
        context = self.base.context(request)
        return render(request, "musiq.html", context)

    def state_dict(self):
        state_dict = self.base.state_dict()
        try:
            current_song = CurrentSong.objects.get()
            current_song = model_to_dict(current_song)
        except CurrentSong.DoesNotExist:
            current_song = None
        song_queue = []
        all_songs = self.queue.all()
        if self.base.settings.voting_system:
            all_songs = all_songs.order_by("-votes", "index")
        for song in all_songs:
            song_dict = model_to_dict(song)
            song_dict["duration_formatted"] = song_utils.format_seconds(
                song_dict["duration"]
            )
            song_dict["confirmed"] = True
            # find the query of the placeholder that this song replaces (if any)
            for i, placeholder in enumerate(self.placeholders[:]):
                if placeholder["replaced_by"] == song.id:
                    song_dict["replaces"] = placeholder["query"]
                    self.placeholders.remove(placeholder)
                    break
            else:
                song_dict["replaces"] = None
            song_queue.append(song_dict)
        song_queue += [
            {"title": placeholder["query"], "confirmed": False}
            for placeholder in self.placeholders
        ]

        if state_dict["alarm"]:
            state_dict["current_song"] = {
                "queue_key": -1,
                "manually_requested": False,
                "votes": None,
                "internal_url": "",
                "external_url": "",
                "artist": "Raveberry",
                "title": "ALARM!",
                "duration": 10,
                "created": "",
            }
        else:
            state_dict["current_song"] = current_song
        state_dict["paused"] = self.player.paused()
        state_dict["progress"] = self.player.progress()
        state_dict["shuffle"] = self.player.shuffle
        state_dict["repeat"] = self.player.repeat
        state_dict["autoplay"] = self.player.autoplay
        state_dict["volume"] = self.player.volume
        state_dict["song_queue"] = song_queue
        return state_dict

    def get_state(self, request):
        state = self.state_dict()
        return JsonResponse(state)

    def update_state(self):
        state_handler.update_state(self.state_dict())



from django.db import transaction
from django.db.models import F
from django.http import HttpResponse
from django.http import HttpResponseForbidden
from django.http import HttpResponseBadRequest
from django.utils import timezone
from django.conf import settings

from core.models import Setting
import core.models as models
import core.musiq.song_utils as song_utils

from threading import Semaphore
from threading import Lock
from threading import Event
from datetime import datetime
from functools import wraps
from contextlib import contextmanager
from requests.exceptions import ConnectionError
import os
import time
import random
import subprocess
import mopidy.core
import mopidy.backend
from mopidyapi import MopidyAPI
from mopidyapi.exceptions import MopidyError

from core.musiq.music_provider import SongProvider
from core.util import background_thread


class Player:
    queue_semaphore = None

    def __init__(self, musiq):
        self.musiq = musiq

        self.SEEK_DISTANCE = 10 * 1000
        self.shuffle = (
            self.musiq.base.settings.get_setting("shuffle", "False") == "True"
        )
        self.repeat = self.musiq.base.settings.get_setting("repeat", "False") == "True"
        self.autoplay = (
            self.musiq.base.settings.get_setting("autoplay", "False") == "True"
        )

        self.queue = models.QueuedSong.objects
        Player.queue_semaphore = Semaphore(self.queue.count())
        self.alarm_playing = Event()
        self.running = True

        self.player = MopidyAPI()
        self.player_lock = Lock()
        with self.mopidy_command(important=True):
            self.player.playback.stop()
            self.player.tracklist.clear()
            # make songs disappear from tracklist after being played
            self.player.tracklist.set_consume(True)

        with self.mopidy_command(important=True):
            # currentsong = self.player.currentsong()
            self.volume = self.player.mixer.get_volume() / 100

    def start(self):
        self._loop()

    def progress(self):
        # the state is either pause or stop
        current_position = 0
        duration = 1
        with self.mopidy_command() as allowed:
            if allowed:
                current_position = self.player.playback.get_time_position()
                current_track = self.player.playback.get_current_track()
                if current_track is None:
                    return 0
                duration = current_track.length
        return 100 * current_position / duration

    def paused(self):
        # the state is either pause or stop
        paused = False
        with self.mopidy_command() as allowed:
            if allowed:
                paused = (
                    self.player.playback.get_state()
                    != mopidy.core.PlaybackState.PLAYING
                )
        return paused

    @background_thread
    def _loop(self):
        while True:

            catch_up = None
            if models.CurrentSong.objects.exists():
                # recover interrupted song from database
                current_song = models.CurrentSong.objects.get()

                # continue with the current song (approximately) where we last left
                song_provider = SongProvider.create(
                    self.musiq, external_url=current_song.external_url
                )
                duration = song_provider.get_metadata()["duration"]
                catch_up = round(
                    (timezone.now() - current_song.created).total_seconds() * 1000
                )
                if catch_up > duration * 1000:
                    catch_up = -1
            else:
                self.queue_semaphore.acquire()
                if not self.running:
                    break

                # select the next song depending on settings
                if self.musiq.base.settings.voting_system:
                    with transaction.atomic():
                        song = self.queue.all().order_by("-votes", "index")[0]
                        song_id = song.id
                        self.queue.remove(song.id)
                elif self.shuffle:
                    index = random.randint(0, models.QueuedSong.objects.count() - 1)
                    song_id = models.QueuedSong.objects.all()[index].id
                    song = self.queue.remove(song_id)
                else:
                    # move the first song in the queue into the current song
                    song_id, song = self.queue.dequeue()

                if song is None:
                    # either the semaphore didn't match up with the actual count of songs in the queue or a race condition occured
                    self.musiq.base.logger.info("dequeued on empty list")
                    continue

                current_song = models.CurrentSong.objects.create(
                    queue_key=song_id,
                    manually_requested=song.manually_requested,
                    votes=song.votes,
                    internal_url=song.internal_url,
                    external_url=song.external_url,
                    artist=song.artist,
                    title=song.title,
                    duration=song.duration,
                )

                self._handle_autoplay()

                try:
                    archived_song = models.ArchivedSong.objects.get(
                        url=current_song.external_url
                    )
                    if self.musiq.base.settings.voting_system:
                        votes = current_song.votes
                    else:
                        votes = None
                    if self.musiq.base.settings.logging_enabled:
                        models.PlayLog.objects.create(
                            song=archived_song,
                            manually_requested=current_song.manually_requested,
                            votes=votes,
                        )
                except (
                    models.ArchivedSong.DoesNotExist,
                    models.ArchivedSong.MultipleObjectsReturned,
                ):
                    pass

            self.musiq.update_state()

            playing = Event()

            @self.player.on_event("playback_state_changed")
            def on_playback_state_changed(event):
                playing.set()

            with self.mopidy_command(important=True):
                # after a restart consume may be set to False again, so make sure it is on
                self.player.tracklist.clear()
                self.player.tracklist.set_consume(True)
                self.player.tracklist.add(uris=[current_song.internal_url])
                self.player.playback.play()
                # mopidy can only seek when the song is playing
                playing.wait(timeout=1)
                if catch_up is not None and catch_up >= 0:
                    self.player.playback.seek(catch_up)

            self.musiq.update_state()

            if catch_up is None or catch_up >= 0:
                if not self._wait_until_song_end():
                    # there was a ConnectionError during waiting for the song to end
                    # thus, we do not delete the current song but recover its state by restarting the loop
                    continue

            current_song.delete()

            if self.repeat:
                song_provider = SongProvider.create(
                    self.musiq, external_url=current_song.external_url
                )
                self.queue.enqueue(song_provider.get_metadata(), False)
                self.queue_semaphore.release()
            else:
                # the song left the queue, we can delete big downloads
                song_utils.decide_deletion(current_song.internal_url)

            self.musiq.update_state()

            if (
                self.musiq.base.user_manager.partymode_enabled()
                and random.random() < self.musiq.base.settings.alarm_probability
            ):
                self.alarm_playing.set()
                self.musiq.base.lights.alarm_started()

                self.musiq.update_state()

                with self.mopidy_command(important=True):
                    self.player.tracklist.add(
                        uris=[
                            "file://"
                            + os.path.join(settings.BASE_DIR, "config/sounds/alarm.m4a")
                        ]
                    )
                    self.player.playback.play()
                playing.clear()
                playing.wait(timeout=1)
                self._wait_until_song_end()

                self.musiq.base.lights.alarm_stopped()
                self.musiq.update_state()
                self.alarm_playing.clear()

    def _wait_until_song_end(self):
        # wait until the song is over. Returns True when finished without errors, False otherwise
        """playback_ended = Event()
        @self.player.on_event('tracklist_changed')
        def on_tracklist_change(event):
            playback_ended.set()
        playback_ended.wait()"""
        error = False
        while True:
            with self.mopidy_command() as allowed:
                if allowed:
                    try:
                        if (
                            self.player.playback.get_state()
                            == mopidy.core.PlaybackState.STOPPED
                        ):
                            break
                    except (ConnectionError, MopidyError) as e:
                        # error during state get, skip until reconnected
                        error = True
            time.sleep(0.1)
        return not error

    def _handle_autoplay(self, url=None):
        if self.autoplay and models.QueuedSong.objects.count() == 0:
            if url is None:
                # if no url was specified, use the one of the current song
                try:
                    current_song = models.CurrentSong.objects.get()
                    url = current_song.external_url
                except (
                    models.CurrentSong.DoesNotExist,
                    models.CurrentSong.MultipleObjectsReturned,
                ):
                    return

            provider = SongProvider.create(self.musiq, external_url=url)
            try:
                suggestion = provider.get_suggestion()
            except Exception as e:
                self.musiq.base.logger.error("error during suggestions for " + url)
                self.musiq.base.logger.error(e)
            else:
                self.musiq._request_music(
                    "",
                    suggestion,
                    None,
                    False,
                    provider.type,
                    archive=False,
                    manually_requested=False,
                )

    # wrapper method for our mopidy client that pings the mopidy server before any command and reconnects if necessary.
    @contextmanager
    def mopidy_command(self, important=False):
        timeout = 3
        if important:
            timeout = -1
        if self.player_lock.acquire(timeout=timeout):
            yield True
            self.player_lock.release()
        else:
            print("mopidy command could not be executed")
            yield False

    # every control changes the views state and returns an empty response
    def control(func):
        def _decorator(self, request, *args, **kwargs):
            # don't allow controls during alarm
            if self.alarm_playing.is_set():
                return HttpResponseBadRequest()
            func(self, request, *args, **kwargs)
            self.musiq.update_state()
            return HttpResponse()

        return wraps(func)(_decorator)

    # in the voting system only the admin can control the player
    def disabled_when_voting(func):
        def _decorator(self, request, *args, **kwargs):
            if (
                self.musiq.base.settings.voting_system
                and not self.musiq.base.user_manager.has_controls(request.user)
            ):
                return HttpResponseForbidden()
            func(self, request, *args, **kwargs)
            self.musiq.update_state()
            return HttpResponse()

        return wraps(func)(_decorator)

    @disabled_when_voting
    @control
    def restart(self, request):
        with self.mopidy_command() as allowed:
            if allowed:
                self.player.playback.seek(0)

    @disabled_when_voting
    @control
    def seek_backward(self, request):
        with self.mopidy_command() as allowed:
            if allowed:
                current_position = self.player.playback.get_time_position()
                self.player.playback.seek(current_position - self.SEEK_DISTANCE)

    @disabled_when_voting
    @control
    def play(self, request):
        # unlikely race condition possible. but its impact is not worth the effort to prevent it
        with self.mopidy_command() as allowed:
            if allowed:
                self.player.playback.play()

    @disabled_when_voting
    @control
    def pause(self, request):
        with self.mopidy_command() as allowed:
            if allowed:
                self.player.playback.pause()

    @disabled_when_voting
    @control
    def seek_forward(self, request):
        with self.mopidy_command() as allowed:
            if allowed:
                current_position = self.player.playback.get_time_position()
                self.player.playback.seek(current_position + self.SEEK_DISTANCE)

    @disabled_when_voting
    @control
    def skip(self, request):
        with self.mopidy_command() as allowed:
            if allowed:
                self.player.playback.next()

    @disabled_when_voting
    @control
    def set_shuffle(self, request):
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="shuffle").update(value=enabled)
        self.shuffle = enabled

    @disabled_when_voting
    @control
    def set_repeat(self, request):
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="repeat").update(value=enabled)
        self.repeat = enabled

    @disabled_when_voting
    @control
    def set_autoplay(self, request):
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="autoplay").update(value=enabled)
        self.autoplay = enabled
        self._handle_autoplay()

    @disabled_when_voting
    @control
    def set_volume(self, request):
        self.volume = float(request.POST.get("value"))
        with self.mopidy_command() as allowed:
            if allowed:
                self.player.mixer.set_volume(round(self.volume * 100))

    @disabled_when_voting
    @control
    def remove_all(self, request):
        if not self.musiq.base.user_manager.is_admin(request.user):
            return HttpResponseForbidden()
        with self.mopidy_command() as allowed:
            if allowed:
                with transaction.atomic():
                    count = self.queue.count()
                    self.queue.all().delete()
                for _ in range(count):
                    self.queue_semaphore.acquire(blocking=False)

    @disabled_when_voting
    @control
    def prioritize(self, request):
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)
        self.queue.prioritize(key)

    @disabled_when_voting
    @control
    def remove(self, request):
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)
        try:
            removed = self.queue.remove(key)
            self.queue_semaphore.acquire(blocking=False)
            # if we removed a song and it was added by autoplay,
            # we want it to be the new basis for autoplay
            if not removed.manually_requested:
                self._handle_autoplay(removed.external_url)
            else:
                self._handle_autoplay()
        except models.QueuedSong.DoesNotExist:
            return HttpResponseBadRequest("song does not exist")

    @disabled_when_voting
    @control
    def reorder(self, request):
        prev = request.POST.get("prev")
        key = request.POST.get("element")
        next = request.POST.get("next")
        if key is None or len(key) == 0:
            return HttpResponseBadRequest()
        if prev is None or len(prev) == 0:
            prev = None
        else:
            prev = int(prev)
        key = int(key)
        if next is None or len(next) == 0:
            next = None
        else:
            next = int(next)
        try:
            self.queue.reorder(prev, key, next)
        except ValueError:
            return HttpResponseBadRequest("request on old state")

    @control
    def vote_up(self, request):
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)

        models.CurrentSong.objects.filter(queue_key=key).update(votes=F("votes") + 1)

        self.queue.vote_up(key)

    @control
    def vote_down(self, request):
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)

        models.CurrentSong.objects.filter(queue_key=key).update(votes=F("votes") - 1)
        try:
            current_song = models.CurrentSong.objects.get()
            if (
                current_song.queue_key == key
                and current_song.votes <= -self.musiq.base.settings.downvotes_to_kick
            ):
                with self.mopidy_command() as allowed:
                    if allowed:
                        self.player.playback.next()
        except models.CurrentSong.DoesNotExist:
            pass

        removed = self.queue.vote_down(key, -self.musiq.base.settings.downvotes_to_kick)
        # if we removed a song by voting, and it was added by autoplay,
        # we want it to be the new basis for autoplay
        if removed is not None:
            self.queue_semaphore.acquire(blocking=False)
            if not removed.manually_requested:
                self._handle_autoplay(removed.external_url)
            else:
                self._handle_autoplay()

    @control
    def start_loop(self, request):
        if not self.musiq.base.user_manager.is_admin(request.user):
            return HttpResponseForbidden()
        if self.running:
            return HttpResponse()
        # start the main loop, only used for tests
        self.running = True
        self.start()

    @control
    def stop_loop(self, request):
        if not self.musiq.base.user_manager.is_admin(request.user):
            return HttpResponseForbidden()
        # stop the main loop, only used for tests
        self.running = False
        self.queue_semaphore.release()



from django.conf import settings

import os
import math
import errno
import colorsys
import subprocess


class VizProgram:
    def __init__(self, lights):
        self.lights = lights
        self.consumers = 0

    def start(self):
        """ initializes the program, allocates resources """
        pass

    def use(self):
        """ tells the program that it is used by another consumer
        starts the program if this was the first one """
        if self.consumers == 0:
            self.start()
        self.consumers += 1

    def compute(self):
        """ is called once per led frame. computation should happen here, 
        results are only returned in color functions"""
        pass

    def stop(self):
        """ stops the program, releases resources """
        pass

    def release(self):
        """ tells the program that one consumer does not use it anymore
        stops the program if this was the last one """
        self.consumers -= 1
        if self.consumers == 0:
            self.stop()


class Disabled(VizProgram):
    def __init__(self, lights):
        super().__init__(lights)
        self.name = "Disabled"

    def ring_colors(self):
        print("requested disabled ring_colors!")

    def strip_color(self):
        print("requested disabled strip_color!")

    def draw(self):
        print("requested disabled draw!")

    def increase_resolution(self):
        pass

    def decrease_resolution(self):
        pass


class Fixed(VizProgram):
    def __init__(self, lights):
        super().__init__(lights)
        self.name = "Fixed"
        self.color = (0, 0, 0)

    def compute(self):
        # show a red color if the alarm is active
        alarm_factor = self.lights.alarm_program.factor
        if alarm_factor != -1:
            self.color = (alarm_factor, 0, 0)

    def ring_colors(self):
        return [self.color for led in range(self.lights.ring.LED_COUNT)]

    def strip_color(self):
        return self.color


class Rainbow(VizProgram):
    def __init__(self, lights):
        super().__init__(lights)
        self.name = "Rainbow"
        self.program_duration = 1

    def start(self):
        self.time_passed = 0

    def compute(self):
        self.time_passed += self.lights.seconds_per_frame * self.lights.program_speed
        self.time_passed %= self.program_duration
        self.current_fraction = self.time_passed / self.program_duration

    def ring_colors(self):
        return [
            colorsys.hsv_to_rgb(
                (self.current_fraction + led / self.lights.ring.LED_COUNT) % 1, 1, 1
            )
            for led in range(self.lights.ring.LED_COUNT)
        ]

    def strip_color(self):
        return colorsys.hsv_to_rgb(self.current_fraction, 1, 1)


class Adaptive(VizProgram):
    def __init__(self, lights):
        super().__init__(lights)
        self.name = "Rave"
        self.cava = self.lights.cava_program

        # RING
        # map the leds to rainbow colors from red over green to blue (without pink-> hue values in [0, ⅔]
        # stretch the outer regions (red and blue) and compress the inner region (green)
        self.led_count = self.lights.ring.LED_COUNT
        hues = [
            (2 / 3)
            * 1
            / (1 + math.e ** (-4 * math.e * (led / (self.led_count - 1) - 0.5)))
            for led in range(0, self.led_count)
        ]
        self.base_colors = [colorsys.hsv_to_rgb(hue, 1, 1) for hue in hues]

        # STRIP
        # distribute frequencies over the three leds. Don't use hard cuts, but smooth functions
        # the functions add up to one at every point and each functions integral is a third
        self.red_coeffs = [
            -1 / (1 + math.e ** (-6 * math.e * (led / (self.led_count - 1) - 1 / 3)))
            + 1
            for led in range(0, self.led_count)
        ]
        self.blue_coeffs = [
            1 / (1 + math.e ** (-6 * math.e * (led / (self.led_count - 1) - 2 / 3)))
            for led in range(0, self.led_count)
        ]
        self.green_coeffs = [
            1 - self.red_coeffs[led] - self.blue_coeffs[led]
            for led in range(0, self.led_count)
        ]

    def start(self):
        self.cava.use()

    def compute(self):
        # aggregate the length of cavas frame into a list the length of the number of leds we have. This reduces computation
        values_per_led = len(self.cava.current_frame) // self.led_count
        self.current_frame = []
        for led in range(self.led_count):
            self.current_frame.append(
                sum(
                    self.cava.current_frame[
                        led * values_per_led : (led + 1) * values_per_led
                    ]
                )
                / values_per_led
            )

    def ring_colors(self):
        return [
            tuple(factor * val for val in color)
            for factor, color in zip(self.current_frame, self.base_colors)
        ]

    def strip_color(self):
        red = (
            sum(coeff * val for coeff, val in zip(self.red_coeffs, self.current_frame))
            * 3
            / self.led_count
        )
        green = (
            sum(
                coeff * val for coeff, val in zip(self.green_coeffs, self.current_frame)
            )
            * 3
            / self.led_count
        )
        blue = (
            sum(coeff * val for coeff, val in zip(self.blue_coeffs, self.current_frame))
            * 3
            / self.led_count
        )
        red = min(1, red)
        green = min(1, green)
        blue = min(1, blue)
        return (red, green, blue)

    def stop(self):
        self.cava.release()


class Alarm(VizProgram):
    def __init__(self, lights):
        super().__init__(lights)
        self.name = "Alarm"
        self.time_passed = 0
        self.sound_count = 0
        self.increasing_duration = 0.45
        self.decreasing_duration = 0.8
        self.sound_duration = 2.1
        self.sound_repetition = 2.5
        self.factor = -1

    def start(self):
        self.time_passed = 0
        self.sound_count = 0
        self.factor = 0

    def compute(self):
        # do not compute if the alarm is not active
        if self.consumers == 0:
            return
        self.time_passed += self.lights.seconds_per_frame
        if self.time_passed >= self.sound_repetition:
            self.sound_count += 1
            self.time_passed %= self.sound_repetition

        if self.sound_count >= 4:
            self.factor = 0
            return
        if self.time_passed < self.increasing_duration:
            self.factor = self.time_passed / self.increasing_duration
        elif self.time_passed < self.sound_duration - self.decreasing_duration:
            self.factor = 1
        elif self.time_passed < self.sound_duration:
            self.factor = (
                1
                - (self.time_passed - (self.sound_duration - self.decreasing_duration))
                / self.decreasing_duration
            )
        else:
            self.factor = 0

    def stop(self):
        self.factor = -1


class Cava(VizProgram):
    def __init__(self, lights):
        super().__init__(lights)

        self.cava_fifo_path = os.path.join(settings.BASE_DIR, "config/cava_fifo")

        # Keep these configurations in sync with config/cava.config
        self.bars = 199
        self.bit_format = 8

        self.frame_length = self.bars * (self.bit_format // 8)

    def start(self):
        self.current_frame = [0 for led in range(self.bars)]
        self.growing_frame = b""
        try:
            # delete old contents of the pipe
            os.remove(self.cava_fifo_path)
        except FileNotFoundError as e:
            # the file does not exist
            pass
        try:
            os.mkfifo(self.cava_fifo_path)
        except FileExistsError as e:
            # the file already exists
            print(self.cava_fifo_path + " already exists while starting")

        self.cava_process = subprocess.Popen(
            ["cava", "-p", os.path.join(settings.BASE_DIR, "config/cava.config")],
            cwd=settings.BASE_DIR,
        )
        # cava_fifo = open(cava_fifo_path, 'r')
        self.cava_fifo = os.open(self.cava_fifo_path, os.O_RDONLY | os.O_NONBLOCK)

    def compute(self):
        # do not compute if no program uses cava
        if self.consumers == 0:
            return
        # read the fifo until we get to the current frame
        while True:
            try:
                read = os.read(
                    self.cava_fifo, self.frame_length - len(self.growing_frame)
                )
                if read == b"":
                    return
                self.growing_frame += read
            except OSError as e:
                if e.errno == errno.EAGAIN or e.errno == errno.EWOULDBLOCK:
                    # there were not enough bytes for a whole frame, keep the old frame
                    return
                else:
                    raise

            # we read a whole frame, update the factors
            if len(self.growing_frame) == self.frame_length:
                # vol = max(0.01, self.lights.base.musiq.player.volume)
                # self.current_frame = [int(b) / 255 / vol for b in self.growing_frame]
                self.current_frame = [int(b) / 255 for b in self.growing_frame]
                self.growing_frame = b""

    def stop(self):
        try:
            os.close(self.cava_fifo)
        except OSError as e:
            print("cava fifo already closed: " + str(e))
        except TypeError as e:
            print("cava fifo does not exist: " + str(e))

        self.cava_process.terminate()

        try:
            os.remove(self.cava_fifo_path)
        except FileNotFoundError as e:
            # the file was already deleted
            print(self.cava_fifo_path + " not found while deleting: " + str(e))



from django.db import models
from django.contrib import admin
import core.musiq.song_queue
import core.musiq.song_utils as song_utils

# Create your models here.
class Tag(models.Model):
    text = models.CharField(max_length=100)

    def __str__(self):
        return self.text


class Counter(models.Model):
    value = models.IntegerField()

    def __str__(self):
        return str(self.value)


class ArchivedSong(models.Model):
    url = models.CharField(max_length=200, unique=True)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    counter = models.IntegerField()

    def __str__(self):
        return self.title + " (" + self.url + "): " + str(self.counter)

    def displayname(self):
        return song_utils.displayname(self.artist, self.title)


class ArchivedPlaylist(models.Model):
    list_id = models.CharField(max_length=200)
    title = models.CharField(max_length=1000)
    created = models.DateTimeField(auto_now_add=True)
    counter = models.IntegerField()

    def __str__(self):
        return self.title + ": " + str(self.counter)


class PlaylistEntry(models.Model):
    playlist = models.ForeignKey(
        "ArchivedPlaylist", on_delete=models.CASCADE, related_name="entries"
    )
    index = models.IntegerField()
    url = models.CharField(max_length=200)

    def __str__(self):
        return self.playlist.title + "[" + str(self.index) + "]: " + self.url

    class Meta:
        ordering = ["playlist", "index"]


class ArchivedQuery(models.Model):
    song = models.ForeignKey(
        "ArchivedSong", on_delete=models.CASCADE, related_name="queries"
    )
    query = models.CharField(max_length=1000)

    def __str__(self):
        return self.query


class ArchivedPlaylistQuery(models.Model):
    playlist = models.ForeignKey(
        "ArchivedPlaylist", on_delete=models.CASCADE, related_name="queries"
    )
    query = models.CharField(max_length=1000)

    def __str__(self):
        return self.query


class QueuedSong(models.Model):
    index = models.IntegerField()
    manually_requested = models.BooleanField()
    votes = models.IntegerField(default=0)
    internal_url = models.CharField(max_length=200)
    external_url = models.CharField(max_length=200, blank=True)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    duration = models.IntegerField()
    objects = core.musiq.song_queue.SongQueue()

    def __str__(self):
        return str(self.index) + ": " + self.title + " (" + self.internal_url + ")"

    def displayname(self):
        return song_utils.displayname(self.artist, self.title)

    class Meta:
        ordering = ["index"]


class CurrentSong(models.Model):
    queue_key = models.IntegerField()
    manually_requested = models.BooleanField()
    votes = models.IntegerField()
    internal_url = models.CharField(max_length=200)
    external_url = models.CharField(max_length=200, blank=True)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    duration = models.IntegerField()
    created = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return self.title + " (" + self.internal_url + ")"

    def displayname(self):
        return song_utils.displayname(self.artist, self.title)


class RequestLog(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    song = models.ForeignKey(
        "ArchivedSong", on_delete=models.SET_NULL, blank=True, null=True
    )
    playlist = models.ForeignKey(
        "ArchivedPlaylist", on_delete=models.SET_NULL, blank=True, null=True
    )
    address = models.CharField(max_length=50)

    def __str__(self):
        if self.song is not None:
            return self.address + ": " + self.song.displayname()
        elif self.playlist is not None:
            return self.address + ": " + self.playlist.title
        else:
            return self.address + ": <None>"


class PlayLog(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    song = models.ForeignKey(
        "ArchivedSong", on_delete=models.SET_NULL, blank=True, null=True
    )
    manually_requested = models.BooleanField()
    votes = models.IntegerField(null=True)

    def __str__(self):
        return (
            "played " + self.song.displayname() + " with " + str(self.votes) + " votes"
        )


class Setting(models.Model):
    key = models.CharField(max_length=200, unique=True)
    value = models.CharField(max_length=200)

    def __str__(self):
        return self.key + ": " + ("None" if self.value is None else self.value)


class Pad(models.Model):
    version = models.IntegerField(default=0)
    content = models.CharField(max_length=100000)

    def __str__(self):
        return "{" + str(self.version) + "}: " + self.content[:20] + "..."



import mopidy_spotify
from django.http import HttpResponse

from core.musiq import song_utils
from core.musiq.music_provider import SongProvider, PlaylistProvider
from core.models import ArchivedSong, Setting
from mopidy_spotify.web import OAuthClient

from urllib.parse import urlparse

from core.models import ArchivedSong, ArchivedPlaylist

_web_client = None


def get_web_client():
    global _web_client
    if _web_client is None:
        client_id = Setting.objects.get(key="spotify_client_id").value
        client_secret = Setting.objects.get(key="spotify_client_secret").value
        _web_client = OAuthClient(
            base_url="https://api.spotify.com/v1",
            refresh_url="https://auth.mopidy.com/spotify/token",
            client_id=client_id,
            client_secret=client_secret,
        )
    return _web_client


def get_search_suggestions(query, playlist):
    web_client = get_web_client()
    result = web_client.get(
        "search",
        params={
            "q": query,
            "limit": "10",
            "market": "from_token",
            "type": "playlist" if playlist else "track",
        },
    )

    if playlist:
        items = result["playlists"]["items"]
    else:
        items = result["tracks"]["items"]

    suggestions = []
    for item in items:
        external_url = item["external_urls"]["spotify"]
        title = item["name"]
        if playlist:
            displayname = title
        else:
            artist = item["artists"][0]["name"]
            displayname = song_utils.displayname(artist, title)
        suggestions.append((displayname, external_url))

    # remove duplicates
    chosen_displaynames = set()
    unique_suggestions = []
    for suggestion in suggestions:
        if suggestion[0] in chosen_displaynames:
            continue
        unique_suggestions.append(suggestion)
        chosen_displaynames.add(suggestion[0])
    return unique_suggestions


class SpotifySongProvider(SongProvider):
    @staticmethod
    def get_id_from_external_url(url):
        return urlparse(url).path.split("/")[-1]

    @staticmethod
    def get_id_from_internal_url(url):
        return url.split(":")[-1]

    def __init__(self, musiq, query, key):
        super().__init__(musiq, query, key)
        self.type = "spotify"
        self.spotify_library = musiq.player.player.library
        self.metadata = dict()
        self.web_client = get_web_client()

    def check_cached(self):
        if self.query is not None and self.query.startswith(
            "https://open.spotify.com/"
        ):
            extracted_id = SpotifySongProvider.get_id_from_external_url(self.query)
            if extracted_id is not None:
                self.id = extracted_id

        if not self._check_cached():
            return False
        # Spotify songs cannot be cached and have to be streamed everytime
        return False

    def check_downloadable(self):
        if self.id is None:
            results = self.spotify_library.search({"any": [self.query]})

            try:
                track_info = results[0].tracks[0]
            except AttributeError:
                self.error = "Song not found"
                return False
            self.id = SpotifySongProvider.get_id_from_internal_url(track_info.uri)
            self.gather_metadata(track_info=track_info)
        else:
            self.gather_metadata()

        return True

    def download(self, ip, background=True, archive=True, manually_requested=True):
        self.enqueue(ip, archive=archive, manually_requested=manually_requested)
        # spotify need to be streamed, no download possible
        return True

    def gather_metadata(self, track_info=None):
        if not track_info:
            results = self.spotify_library.search({"uri": [self.get_internal_url()]})
            track_info = results[0].tracks[0]

        self.metadata["internal_url"] = track_info.uri
        self.metadata["external_url"] = self.get_external_url()
        self.metadata["artist"] = track_info.artists[0].name
        self.metadata["title"] = track_info.name
        self.metadata["duration"] = track_info.length / 1000

    def get_metadata(self):
        if not self.metadata:
            self.gather_metadata()
        return self.metadata

    def get_internal_url(self):
        return "spotify:track:" + self.id

    def get_external_url(self):
        return "https://open.spotify.com/track/" + self.id

    def get_suggestion(self):
        result = self.web_client.get(
            "recommendations",
            params={"limit": "1", "market": "from_token", "seed_tracks": self.id,},
        )

        try:
            external_url = result["tracks"][0]["external_urls"]["spotify"]
        except IndexError:
            self.error = "no recommendation found"
            return None

        return external_url

    def request_radio(self, ip):
        result = self.web_client.get(
            "recommendations",
            params={
                "limit": self.musiq.base.settings.max_playlist_items,
                "market": "from_token",
                "seed_tracks": self.id,
            },
        )

        for track in result["tracks"]:
            external_url = track["external_urls"]["spotify"]
            self.musiq._request_music(
                "",
                external_url,
                None,
                False,
                "spotify",
                archive=False,
                manually_requested=False,
            )

        return HttpResponse("queueing radio")


class SpotifyPlaylistProvider(PlaylistProvider):
    @staticmethod
    def get_id_from_external_url(url):
        if not url.startswith("https://open.spotify.com/playlist/"):
            return None
        return urlparse(url).path.split("/")[-1]

    def __init__(self, musiq, query, key):
        super().__init__(musiq, query, key)
        self.type = "spotify"
        self.web_client = get_web_client()

    def search_id(self):
        result = self.web_client.get(
            "search",
            params={
                "q": self.query,
                "limit": "1",
                "market": "from_token",
                "type": "playlist",
            },
        )

        try:
            list_info = result["playlists"]["items"][0]
        except IndexError:
            self.error = "No playlist found"
            return None

        list_id = list_info["id"]
        self.title = list_info["name"]

        return list_id

    def is_radio(self):
        return False

    def fetch_metadata(self):
        if self.title is None:
            result = self.web_client.get(
                f"playlists/{self.id}", params={"fields": "name", "limit": "50",},
            )
            self.title = result["name"]

        # download at most 50 tracks for a playlist (spotifys maximum)
        # for more tracks paging would need to be implemented
        result = self.web_client.get(
            f"playlists/{self.id}/tracks",
            params={
                "fields": "items(track(external_urls(spotify)))",
                "limit": "50",
                "market": "from_token",
            },
        )

        track_infos = result["items"]
        for track_info in track_infos:
            self.urls.append(track_info["track"]["external_urls"]["spotify"])



import Adafruit_PCA9685


class Strip:
    def __init__(self):
        try:
            self.controller = Adafruit_PCA9685.PCA9685()
            self.initialized = True
        except (RuntimeError, OSError):
            # LED strip is not connected
            self.initialized = False

        self.brightness = 1

    def set_color(self, color):
        if not self.initialized:
            return

        for channel, val in enumerate(color):
            # map the value to the interval [0, 4095]
            dimmed_val = val * self.brightness
            scaled_val = round(dimmed_val * 4095)
            self.controller.set_pwm(channel, 0, scaled_val)

    def clear(self):
        if not self.initialized:
            return

        for channel in range(3):
            self.controller.set_pwm(channel, 0, 0)



from channels.routing import ProtocolTypeRouter, URLRouter
from channels.auth import AuthMiddlewareStack
import core.routing

application = ProtocolTypeRouter(
    {"websocket": AuthMiddlewareStack(URLRouter(core.routing.websocket_urlpatterns)),}
)



from threading import Thread

from django.db import connection


def background_thread(function):
    def decorator(*args, **kwargs):
        def run_and_close_connection():
            function(*args, **kwargs)
            connection.close()

        t = Thread(target=run_and_close_connection, daemon=True)
        t.start()
        return t

    return decorator



import os
import random

from django.http import HttpResponseBadRequest

from core.models import PlaylistEntry
from core.musiq import song_utils
from core.musiq.music_provider import SongProvider, PlaylistProvider
from main import settings


class LocalSongProvider(SongProvider):
    """
    A class handling local files on the drive.
    If the library is at /home/pi/Music/ and SONGS_CACHE_DIR is at /home/pi/raveberry
    there will be a symlink /home/pi/raveberry/local_library to /home/pi/Music

    Example values for a file at /home/pi/Music/Artist/Title.mp3 are:
    id: Artist/Title.mp3
    external_url: local_library/Artist/Title.mp3
    internal_url: file:///home/pi/local_library/Artist/Title.mp3
    """

    @staticmethod
    def get_id_from_external_url(url):
        return url[len("local_library/") :]

    def __init__(self, musiq, query, key):
        super().__init__(musiq, query, key)
        self.type = "local"

    def check_cached(self):
        if not self._check_cached():
            return False
        return os.path.isfile(self.get_path())

    def check_downloadable(self):
        # Local files can not be downloaded from the internet
        self.error = "Local file missing"
        return False

    def get_metadata(self):
        metadata = song_utils.get_metadata(self.get_path())

        metadata["internal_url"] = self.get_internal_url()
        metadata["external_url"] = self.get_external_url()
        if not metadata["title"]:
            metadata["title"] = metadata["external_url"]

        return metadata

    def get_path(self):
        path = os.path.join(settings.SONGS_CACHE_DIR, self.get_external_url())
        path = path.replace("~", os.environ["HOME"])
        path = os.path.abspath(path)
        return path

    def get_internal_url(self):
        return "file://" + self.get_path()

    def get_external_url(self):
        return "local_library/" + self.id

    def _get_corresponding_playlist(self):
        entries = PlaylistEntry.objects.filter(url=self.get_external_url())
        if not entries.exists():
            raise PlaylistEntry.DoesNotExist()
        # There should be only one playlist containing this song. If there are more, choose any
        index = random.randint(0, entries.count() - 1)
        entry = entries.all()[index]
        playlist = entry.playlist
        return playlist

    def get_suggestion(self):
        playlist = self._get_corresponding_playlist()
        entries = playlist.entries
        index = random.randint(0, entries.count() - 1)
        entry = entries.all()[index]
        return entry.url

    def request_radio(self, ip):
        playlist = self._get_corresponding_playlist()
        return self.musiq._request_music(
            ip,
            playlist.title,
            playlist.id,
            True,
            "local",
            archive=False,
            manually_requested=False,
        )


class LocalPlaylistProvider(PlaylistProvider):
    @staticmethod
    def get_id_from_external_url(url):
        return url[len("local_library/") :]

    def __init__(self, musiq, query, key):
        super().__init__(musiq, query, key)
        self.type = "local"

    def search_id(self):
        self.error = "local playlists can not be downloaded"
        return None

    def is_radio(self):
        return False
